"""Main pyautogit manager class and entry point

The main driver class contains code for common actions performed by all subscreens such as
credential management, as well as functions for switching between subscreens.

Author: Jakub Wlodek  
Created: 01-Oct-2019
"""

# Core Python Utilities
import argparse
import getpass
import json
import os
import shutil
import subprocess
import threading
import datetime
from subprocess import Popen, PIPE

# py_cui library used for Command Line UI construction
import py_cui

# Subscreens and pyautogit modules
import pyautogit.logger as LOGGER
import pyautogit.repo_select_screen as SELECT
import pyautogit.repo_control_screen as CONTROL
import pyautogit.internal_editor_screen as EDITOR
import pyautogit.settings_screen as SETTINGS
import pyautogit.metadata_manager as METADATA


# Module version + copyright
__version__     = '0.0.5'
__copyright__   = '2019-2020'


# Helper pyautogit functions

def find_repos_in_path(path):
    """Helper function that finds repositories in the path

    Parameters
    ----------
    path : str
        Target path
    
    Returns
    -------
    repos : list of str
        list of git repositories within target
    """

    repos = []
    for dir in os.listdir(path):
        new_dir = os.path.join(path, dir)
        if os.path.isdir(new_dir) and is_git_repo(new_dir):
            repos.append(os.path.basename(new_dir))

    LOGGER.write('Found repos in path: {}'.format(repos))

    return repos


def is_git_repo(path):
    """Simple function that checks if a given path is a git repository

    Parameters
    ----------
    path : str
        path to check
    
    Returns
    -------
    is_repo : bool
        True if .git exists, False otherwise
    """

    is_repo = False
    if os.path.exists(os.path.join(path, '.git')):
        is_repo = True
    return is_repo


def parse_args():
    """Function that parses user arguments for pyautogit

    Returns
    -------
    target_repo : str
        The target path for pyautogit
    save_metadata : bool
        flag to say if metadata should be saved
    credentials : list of str
        username, password, if entered
    """

    target_repo = '.'
    credentials = []

    parser = argparse.ArgumentParser(description="A command line interface for git commands.")
    parser.add_argument('-c', '--credentials',      action='store_true', help='Allows user to enter credentials once when pyautogit is started.')
    parser.add_argument('-w', '--workspace',        help='Pass a path to this argument to start pyautogit in a workspace not the current directory.')
    parser.add_argument('-n', '--nosavemetadata',   action='store_true', help='Add this flag if you would like pyautogit to not save metadata between sessions.')
    parser.add_argument('-d', '--debug',            action='store_true', help='Flag that enables debug logging by default.')
    parser.add_argument('-v', '--version',          action='store_true', help='Run pyautogit with this flag to print version information.')
    args = vars(parser.parse_args())

    if args['version']:
        print('pyautogit v{}\n'.format(__version__))
        print('BSD-3-Clause License')
        print('Copyright (c) {} Jakub Wlodek'.format(__copyright__))
        print('https://github.com/jwlodek/pyautogit\n')
        exit()

    if args['credentials']:
        user = input('Please enter your github/gitlab username > ')
        credentials.append(user)
        passwd = getpass.getpass(prompt="Please enter your github/gitlab password > ")
        credentials.append(passwd)
    if args['workspace'] is not None:
        if os.path.exists(args['workspace']):
            if os.path.isdir(args['workspace']):
                os.chdir(args['workspace'])
            else:
                print('ERROR - Path {} is not a directory.'.format(args['workspace']))
                exit(-1)
        else:
            print('ERROR - Path {} does not exist.'.format(args['workspace']))
            exit(-1)

    return target_repo, credentials, args


def main():
    """Entry point for pyautogit. Parses arguments, and initializes the CUI
    """

    target, credentials, args = parse_args()
    save_metadata = not args['nosavemetadata']
    debug_logging = args['debug']

    target_abs = os.path.abspath(target)

    input_type = 'repo'
    if not is_git_repo(target):
        input_type = 'workspace'

    # Make sure we have write permissions
    if not os.access(target, os.W_OK):
        print('ERROR - Permission error for target {}'.format(target_abs))
        exit(-1)
    if input_type == 'repo' and not os.access(os.path.dirname(target_abs), os.W_OK):
        print('ERROR - Permission denied for parent workspace {} of repository {}'.format(os.path.dirname(target_abs), target_abs))
        exit(-1)

    if debug_logging:
        if input_type=='repo':
            LOGGER.set_log_file_path('../.pyautogit/{}.log'.format(str(datetime.datetime.today()).split(' ')[0]))
        else:
            LOGGER.set_log_file_path('.pyautogit/{}.log'.format(str(datetime.datetime.today()).split(' ')[0]))
        LOGGER.toggle_logging()
        LOGGER.write('Initialized debug logging')

    root = py_cui.PyCUI(5, 4)
    
    if debug_logging:
        root.enable_logging(log_file_path='.pyautogit/py_cui_root.log')

    # Use feature added in py_cui 0.0.3 to add unicode widget borders
    root.toggle_unicode_borders()

    _ = PyAutogitManager(root, target, input_type, save_metadata, credentials)
    
    LOGGER.write('Parsed args. Target location - {}'.format(target_abs))
    LOGGER.write('Initial state - {}'.format(input_type))
    LOGGER.write('Initialized manager object, starting CUI...')

    root.start()


# Main pyautogit mananger class


class PyAutogitManager:
    """Main pyautogit manager class. Controls all operations of CUI

    Attributes
    ----------
    root : PyCUI
        The root py_cui window
    target_path : str
        The path to the workspace directory
    save_metadata : bool
        Flag to specify metadata saving or not
    credentials : list of str
        Username and Password for git remote
    current_state : str
        Current state of pyautogit (repo control or repo select)
    default_editor : str
        Command to open external editor
    post_input_callback : no-arg or lambda function
        Function fired after a user input event
    operation_thread : Thread
        A thread for performing async operations. Starts as None, Thread created as needed
    repos : list of str
        List of repositories found in workspace
    repo_select_widget_set : py_cui.widget_set.WidgetSet
        set of py_cui widgets that are parts of the repo select screen
    repo_menu : py_cui.widgets.ScrollMenu
        The repository select menu in the repo select screen
    git_status_box : py_cui.widgets.ScrolledTextBlock
        Main info panel for repo select screen
    current_status_box : py_cui.widgets.ScrolledTextBlock
        Secondary info panel for repo select screen
    clone_new_box : py_cui.widgets.TextBox
        Textbox for cloning new repositories
    create_new_box : py_cui.widgets.TextBox
        Textbox for creating new repositories
    repo_select_manager : RepoSelectManager
        The manager wrapper class for the repo select screen
    repo_control_widget_set : py_cui.widget_set.WidgetSet
        set of py_cui widgets that are parts of the repo control screen
    add_files_menu : py_cui.widgets.ScrollMenu
        Menu for adding/unstaging files
    remotes_menu : py_cui.widgets.ScrollMenu
        Menu for selecting remotes
    branch_menu : py_cui.widgets.ScrollMenu
        Menu for selecting branches
    commits_menu : py_cui.widgets.ScrollMenu
        Menu listing most recent commits
    info_text_block : py_cui.widgets.ScrolledTextBlock
        Main Info text block in repo control screen
    new_branch_textbox : py_cui.widgets.TextBox
        Textbox for creating new branches
    commit_message_box : py_cui.widgets.TextBox
        Textbox for entering new commit messages
    repo_control_manager : RepoControlManager
        Manager wrapper for repo control screen
    """

    def __init__(self, root, target_path, current_state, save_metadata, credentials):
        """Constructor for PyAutogitManager
        """

        self.root = root
        self.repo_select_manager    = SELECT.RepoSelectManager(self)
        self.repo_control_manager   = CONTROL.RepoControlManager(self)
        self.settings_manager       = SETTINGS.SettingsScreen(self)
        self.editor_manager         = EDITOR.EditorScreenManager(self, target_path)
        LOGGER.write('Initialized subscreen managers.')

        self.save_metadata = save_metadata

        # Make sure to convert to absolute path. If we opened a repository, the top path will be one level higher
        self.current_path   = os.path.abspath(target_path)
        self.workspace_path = self.current_path
        if current_state == 'repo':
            self.workspace_path = os.path.dirname(self.current_path)
        
        # Setup some helper objects and default information/variables
        self.current_state  = current_state
        self.default_editor = None
        self.editor_type    = 'Internal'
        
        self.metadata_manager   = METADATA.PyAutogitMetadataManager(self)
        self.loaded_metadata    = self.metadata_manager.read_metadata()
        LOGGER.write('Loaded metadata')
        LOGGER.write(self.loaded_metadata)
        self.metadata_manager.apply_metadata(self.loaded_metadata)
        LOGGER.write('Applied metadata')

        # Stores currently entered credentials
        self.credentials = credentials

        # Temp variable fired on callback after inout
        self.post_input_callback = None

        # Add a run on exit callback to save metadata and close log file
        self.root.run_on_exit(self.close_cleanup)

        # Utility variable used to store user input for callbacks
        self.user_message = None

        # Thread used to perform longer operations
        self.operation_thread = None

        # Repository select screen widgets, key commands.
        self.repos = find_repos_in_path(self.workspace_path)

        # Initialize CUI elements for each sub-screen
        self.repo_select_widget_set     = self.repo_select_manager.initialize_screen_elements()
        self.repo_control_widget_set    = self.repo_control_manager.initialize_screen_elements()
        self.settings_widget_set        = self.settings_manager.initialize_screen_elements()
        self.editor_widget_set          = self.editor_manager.initialize_screen_elements()
        LOGGER.write('Initialized CUI elements')

        # Open repo select screen in workspace view
        if self.current_state == 'workspace':
            self.open_repo_select_window()

        # Open repo control screen in repo view
        elif self.current_state == 'repo':
            self.open_autogit_window()


    def close_cleanup(self):
        """Function fired upon closing pyautogit
        """

        if self.save_metadata:
            self.metadata_manager.write_metadata()
        LOGGER.close_logger()


    def clean_exit(self):
        """Function that exits the CUI cleanly
        """

        LOGGER.write('Exiting pyautogit.')
        self.close_cleanup()
        exit()


    def error_exit(self):
        """Function that exits the CUI with an error code
        """

        LOGGER.write('Exiting with error!')
        self.close_cleanup()
        exit(-1)


    def open_not_supported_popup(self, operation):
        """Function that displays warning for a non-supported operation

        Parameters
        ----------
        operation : str
            The name of the non-supported operation
        """

        self.root.show_warning_popup('Warning - Not Supported', 'The {} operation is not yet supported.'.format(operation))


    def open_autogit_window(self):
        """Function that opens the repository control window.
        """

        LOGGER.write('Opening repo control window')
        target = self.repo_select_manager.repo_menu.get()
        self.repo_select_manager.clear_elements()
        self.repo_control_manager.set_initial_values()
        
        self.root.apply_widget_set(self.repo_control_widget_set)
        if self.current_state == 'workspace':
            os.chdir(target)
        self.current_state = 'repo'
        self.root.set_title('pyautogit v{} - {}'.format(__version__, target))
        self.repo_control_manager.refresh_status()


    def open_autogit_window_target(self):
        """Function that opens a repo control window given a target location
        """

        LOGGER.write('Opening autogit control window on target dir.')
        self.repo_select_manager.clear_elements()
        self.repo_control_manager.set_initial_values()
        self.root.apply_widget_set(self.repo_control_widget_set)
        self.repo_control_manager.refresh_status()


    def open_repo_select_window(self):
        """Opens the repo select window. Fired when the backspace key is pressed in the repo control window
        """

        LOGGER.write('Opening repo select window')
        self.repo_control_manager.clear_elements()
        self.settings_manager.clear_elements()
        self.repo_select_manager.set_initial_values()
        
        self.root.apply_widget_set(self.repo_select_widget_set)
        if self.current_state == 'repo':
            os.chdir('..')
        self.current_state = 'workspace'
        self.root.set_title('pyautogit v{} - {}'.format(__version__, os.path.basename(os.getcwd())))
        self.repo_select_manager.refresh_status()
        self.root.move_focus(self.repo_select_manager.repo_menu)


    def open_settings_window(self):
        """Function for opening the settings window
        """

        LOGGER.write('Opening settings window')
        self.repo_select_manager.clear_elements()
        self.settings_manager.set_initial_values()
        self.root.apply_widget_set(self.settings_widget_set)
        self.root.set_title('pyautogit v{} Settings'.format(__version__))
        self.current_state = 'settings'
        self.settings_manager.refresh_status()


    def open_editor_window(self):
        """Function that opens an editor window
        """

        LOGGER.write('Opening Editor Window')
        self.editor_manager.open_new_directory_external(os.getcwd())
        self.editor_manager.set_initial_values()
        self.root.apply_widget_set(self.editor_widget_set)
        self.current_state == 'editor'
        self.editor_manager.refresh_status()
        self.root.move_focus(self.editor_manager.file_menu)
    

    #-------------------------------------------
    # Credential handler functions
    #-------------------------------------------


    def update_password(self, passwd):
        """Function called once password is entered. 

        If necessary, fires the post_input_callback function

        Parameters
        ----------
        passwd : str
            The user's password
        """

        self.credentials.append(passwd)
        self.repo_select_manager.refresh_status()
        LOGGER.write('User credentials entered')
        if self.post_input_callback is not None:
            self.post_input_callback()
        self.post_input_callback = None


    def ask_password(self, user):
        """Function that opens popup and asks for password. Also writes username to credentials.

        Parameters
        ----------
        user : str
            The user's username
        """

        self.credentials.append(user)
        self.root.show_text_box_popup('Please enter your git remote password', self.update_password, password=True)


    def ask_credentials(self, callback=None):
        """Function that asks for user credentials and places them in the appropriate variables.

        Parameters
        ----------
        callback : function
            Default None, otherwise function called after credentials are entered.
        """

        del self.credentials[:]
        if callback is not None:
            self.post_input_callback = callback
        self.root.show_text_box_popup('Please enter your git remote username', self.ask_password)



    def were_credentials_entered(self):
        """Simple function for checking if credentials were entered

        Returns
        -------
        were_credentials_entered : bool
            True if credentials found, otherwise false
        """

        return len(self.credentials) == 2


    def perform_long_operation(self, title, long_operation_function, post_loading_callback):
        """Function that wraps an operation around a loading icon popup.

        Parameters
        ----------
        title : str
            title for loading icon
        long_operation_function : function
            operation to perform in the background
        post_loading_callback : function
            Function fired once long operation is finished.
        """

        LOGGER.write('Executing long operation {}'.format(title))
        self.root.show_loading_icon_popup('Please Wait', title, callback = post_loading_callback)
        self.operation_thread = threading.Thread(target=long_operation_function)
        self.operation_thread.start()


    def update_default_editor(self):
        """Function that sets the default editor

        Parameters
        ----------
        editor : str
            command line call to open the editor
        """

        LOGGER.write('Updating the default editor to {}'.format(self.user_message))
        self.default_editor = self.user_message
        self.editor_type = 'External'
        self.root.show_message_popup('Default Editor Changed', '{} editor will be used to open directories'.format(self.user_message))
        self.repo_select_manager.refresh_status()


    def ask_default_editor(self):
        """Function that asks user to enter a default text editor
        """

        self.ask_message('Please enter a default editor command. (Ex. code, emacs)', callback=self.update_default_editor)


    def update_message(self, message):
        """Function that is run after user inputs message
        
        Parameters
        ----------
        message : str
            User returned input
        """

        self.user_message = message
        if self.post_input_callback is not None:
            self.post_input_callback()
        self.post_input_callback = None


    def ask_message(self, prompt, callback=None):
        """Function that asks the user for input.
        
        Parameters
        ----------
        prompt : str
            Prompt for user input
        callback : function
            Default None, otherwise, function fired after credentials are asked
        """

        if callback is not None:
            self.post_input_callback = callback
        self.root.show_text_box_popup(prompt, self.update_message)


    def get_logo_text(self):
        """Generates ascii-art version of pyautogit logo

        Returns
        -------
        logo : str
            ascii-art logo
        """

        logo =         '        _    _ _______ ____   _____ _____ _______\n'
        logo = logo +  '    /\\  | |  | |__   __/ __ \\ / ____|_   _|__   __|\n'
        logo = logo +  '   /  \\ | |  | |  | | | |  | | |  __  | |    | |   \n'
        logo = logo +  '  / /\\ \\| |  | |  | | | |  | | | |_ | | |    | |   \n'
        logo = logo +  ' / ____ \\ |__| |  | | | |__| | |__| |_| |_   | |   \n'
        logo = logo +  '/_/    \\_\\____/   |_|  \\____/ \\_____|_____|  |_|   \n'
        return logo


    def get_about_info(self, with_logo=True):
        """Generates some about me information

        Parameters
        ----------
        with_logo : bool
            flag to show logo or not.
        
        Returns
        -------
        about_info : str
            string with about information
        """

        if with_logo:
            about_info = ' ' + self.get_logo_text() + '\n\n\n'
        else:
            about_info = '\n'
        about_info = about_info + 'Author: Jakub Wlodek\n\nPython CUI git client: https://github.com/jwlodek/pyautogit\n\n\n'
        about_info = about_info + 'Powered by the py_cui Python Command Line UI library:\n\n'
        about_info = about_info + 'https://github.com/jwlodek/py_cui\n\n\n'
        about_info = about_info + 'Documentation available here:\n\n'
        about_info = about_info + 'pyautogit: https://jwlodek.github.io/pyautogit-docs\n'
        about_info = about_info + 'py_cui:    https://jwlodek.github.io/py_cui-docs\n\n\n'
        about_info = about_info + 'Star me on Github!\n\n'
        about_info = about_info + 'Copyright (c) {} Jakub Wlodek'.format(__copyright__)
        return about_info


    def get_welcome_message(self):
        """Function that gets a basic welcome message shown at first run
        
        Returns
        -------
        welcome : str
            welcome message string
        """

        welcome = '\nWelcome to pyautogit!\n\nThis is a command line interface for working with git projects.\n'
        welcome = welcome + '\nTo begin, take a look at the repositories list to the right.\nThis shows all detected '
        welcome = welcome + 'repositories in the workspace.\nYou may create new repositories, or clone them also to the right.\n'
        welcome = welcome + '\nFor pushing, pulling, and cloning, you will need to enter credentials.\nDo this by pressing "c".\n'
        welcome = welcome + '\nAlso, to set a default editor (that can open files and dirs),\npress "e".\n'
        welcome = welcome + '\nIf you encounter issues, please make a ticket on the github page,\nand if you enjoy pyautogit,'
        welcome = welcome + 'feel free to give\nit a star or a sponsorship.\n\nIf you would like to contribute, feel free to do so as well!'
        return welcome



"""
pyautogit.askpass module

This is an internal module meant to be used with pyautogit for git askpass git credential asking.

The two files are included to differentiate windows and linux, where on linux we use the python3 shebang, on windows the python shebang

(Hopefully with the death of python 2 we can only use one file)
"""


"""File containing class that acts as a parent class to all sub-screen managers

The base screen manager is responsible for defining how CUI elements are assigned to screens,
as well as what the screen must do to refresh its status. Also supports performing long (async)
operations, and running custom commands
"""

from sys import platform
import pyautogit.commands
import pyautogit.logger as LOGGER


class ScreenManager:
    """Main parent screen manager class.

    Contains common functionality for showing command results, handling credentials, commands, and long operations.

    Attributes
    ----------
    manager : PyAutogitManager
        Driver engine manager
    message : str
        A variable to store messages accross functions
    status : int
        A variable to store status codes accross functions
    utility_var : obj
        A variable that can be used to store any data across functions
    menu_choices : list of str
        Overriden by children, list of options that pop up in menu
    info_panel : py_cui.widgets.TextBlock
        The main textblock on the screen, used to display status information.
    """

    def __init__(self, top_manager, screen_type):
        """Constructor for ScreenManager
        """

        self.screen_type = screen_type
        self.manager = top_manager
        self.message = ''
        self.status = 0
        self.utility_var = None
        self.menu_choices = ['About', 'Exit']
        self.info_panel = None


    def initialize_screen_elements(self):
        """Function that must be overridden by subscreen. Creates py_cui_widgets, returns widget set object.
        """

        pass


    def process_menu_selection(self, selection):
        """Processes based on selection returned from the menu

        Parameters
        ----------
        selection : str
            An element of the self.menu_choices list selected by user
        """

        if selection == 'About' and self.info_panel is not None:
            self.info_panel.set_text(self.manager.get_about_info())
        elif selection == 'Exit':
            self.manager.close_cleanup()
            exit()


    def show_menu(self):
        """Opens the menu using the menu item list for screen manager instance
        """
        
        LOGGER.write('Opening {} screen menu'.format(self.screen_type))
        self.manager.root.show_menu_popup('Full Control Menu', self.menu_choices, self.process_menu_selection)


    def show_command_result(self, out, err, show_on_success = True, command_name='Command', success_message='Success', error_message='Error'):
        """Function that displays the result of stdout/err for an external command.

        Parameters
        ----------
        out : str
            stdout string from command
        err : str
            stderr string from command
        show_on_success : bool
            Set to false to show no messages on success. (ex. git log doesnt need success message)
        command_name : str
            name of command run.
        success_message : str
            message to show on successful completion
        error_message : str
            message to show on unsuccessful completion
        """

        show_in_box = False
        stripped_output = out.strip()
        if len(out.splitlines()) > 1:
            popup_message = "Check Info Box For {} Output".format(command_name)
            show_in_box = True
        else:
            popup_message = stripped_output
        if err != 0:
            self.manager.root.show_error_popup(error_message, popup_message)
        elif show_on_success:
            self.manager.root.show_message_popup(success_message, popup_message)
        if show_in_box and (err != 0 or show_on_success):
            box_out = out
            if err != 0:
                err_out = '\n'
                temp = out.splitlines()
                for line in temp:
                    err_out = err_out + '- ' + line + '\n'
                box_out = err_out
            self.info_panel.title = '{} Output'.format(command_name)
            self.info_panel.set_text(box_out)


    def show_status_long_op(self, name='Command', succ_message="Success", err_message = "Error"):
        """Shows the status of a long(async) operation on success completion

        Parameters
        ----------
        name : str
            name of command run.
        succ_message : str
            message to show on successful completion
        err_message : str
            message to show on unsuccessful completion
        """

        self.show_command_result(self.message, self.status, command_name=name, success_message=succ_message, error_message=err_message)
        self.message = ''
        self.status = 0


    def refresh_status(self):
        """Function that is fired after each git operation. Implement in subclasses.
        """

        pass


    def clear_elements(self):
        """Function that clears entries from widgets for reuse
        """

        pass


    def set_initial_values(self):
        """Function that sets initial values for widgets in screen
        """

        pass


    def handle_user_command(self, command):
        """Handles custom user command.

        Parameters
        ----------
        command : str
            The string command entered by the user
        """

        out, err = pyautogit.commands.handle_custom_command(command)
        self.show_command_result(out, err, command_name=command)
        self.refresh_status()


    def ask_custom_command(self):
        """Function that prompts user to enter custom command
        """

        shell='Bash'
        if platform == 'win32':
            shell='Batch'
        self.manager.root.show_text_box_popup('Please Enter A {} Command:'.format(shell), self.handle_user_command)

    
    def execute_long_operation(self, loading_messge, long_op_function, credentials_required=False):
        """Wrapper function that allows for executing long operations w/ credential requirements.

        Parameters
        ----------
        loading_message : str
            Message displayed while async op is performed
        long_op_function : no-arg or lambda function
            Function that is fired in an async second thread
        credentials_required : bool
            If true, prompts to enter credentials before starting async op
        """
        
        if credentials_required and not self.manager.were_credentials_entered():
            self.manager.ask_credentials(callback=lambda : self.manager.perform_long_operation(loading_messge, long_op_function, self.show_status_long_op))
        else:
            self.manager.perform_long_operation(loading_messge, long_op_function, self.show_status_long_op)


"""Metadata Management Classes and Functions
"""

import os
import json
import shutil
import pyautogit
import pyautogit.logger as LOGGER

class PyAutogitMetadataManager:
    """Helper class for managing inter-use metadata for pyautogit

    Attributes
    ----------
    manager : PyAutogitManager
        The top level program manager object
    first_time : bool
        Flag that tells metadata manager if metadata exists
    """

    def __init__(self, manager):
        """Constructor for PyAutogitMetadataManager
        """

        self.manager = manager
        self.first_time = False


    def write_metadata(self):
        """Writes metadata file with cached settings
        """

        settings_dir = os.path.join(self.manager.workspace_path, '.pyautogit')
        settings_file = os.path.join(settings_dir, 'pyautogit_settings.json')
        if not os.path.exists(settings_dir):
            os.mkdir(settings_dir)
        if os.path.exists(settings_file):
            os.remove(settings_file)
        metadata = {}
        metadata['EDITOR']      = self.manager.default_editor
        metadata['VERSION']     = pyautogit.__version__
        metadata['LOG_ENABLE']  = LOGGER._LOG_ENABLED
        LOGGER.write('Writing metadata: {}'.format(metadata))
        fp = open(settings_file, 'w')
        json.dump(metadata, fp)
        fp.close()


    def apply_metadata(self, metadata):
        """Applies metadata from cached settings

        Parameters
        ----------
        metadata : dict
            Metadata parsed from json to python dict.
        """
        if metadata is None:
            return
        if 'EDITOR' in metadata.keys():
            self.manager.default_editor = metadata['EDITOR']
            self.manager.editor_type = 'External'
        if 'VERSION' in metadata.keys() and metadata['VERSION'] != pyautogit.__version__:
            self.manager.root.show_message_popup('PyAutogit Updated', 'Congratulations for updating to pyautogit {}! See patch notes on github.'.format(pyautogit.__version__))
        if 'LOG_ENABLE' in metadata.keys() and metadata['LOG_ENABLE']:
            #LOGGER.toggle_logging()
            pass


    def read_metadata(self):
        """Converts metadata json file to python dict

        Returns
        -------
        metadata : dict
            metadata dictionary
        """

        settings_dir = os.path.join(self.manager.workspace_path, '.pyautogit')
        settings_file = os.path.join(settings_dir, 'pyautogit_settings.json')
        if os.path.exists(settings_file):
            try:
                fp = open(settings_file, 'r')
                metadata = json.load(fp)
                fp.close()
                LOGGER.write('Read metadata:{}'.format(metadata))
                return metadata
            except json.decoder.JSONDecodeError:
                shutil.rmtree(settings_dir)
                self.first_time = True
        else:
            self.first_time = True
            



"""File containing all definitions and functions for git commands used by pyautogit.

This file should remain separate from the CUI interface.

Author: Jakub Wlodek  
Created: 01-Oct-2019
"""

import os
from os import environ
import re
import shutil
import stat
from sys import platform
from subprocess import Popen, PIPE
import pyautogit.askpass as ASKPASS
import pyautogit.logger as LOGGER


def remove_repo_tree(target):
    """Function that removes repository.

    Required since removing git repos on windows require a chmod operation.

    Parameters
    ----------
    target : str
        Dir path to removed repo
    """

    if os.path.exists(target) and os.path.isdir(target):
        def del_rw(action, name, exc):
            os.chmod(name, stat.S_IWRITE)
            os.remove(name)
        shutil.rmtree(target, onerror=del_rw)


def handle_credential_command(command, credentials, target_location='.'):
    """Function that executes a git command that requires credentials.

    Parameters
    ----------
    command : str
        String command to run
    credentials : list of str
        The user's entered git remote credentials
    target_location : str
        Location of repository
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    out = ''
    err = 0
    # This is a bit janky, but I'm not sure what I could do to make it better
    #askpass_dir = os.path.dirname(ASKPASS.__file__)
    if platform == "win32":
        askpass_script = "askpass_pyautogit_win"
    else:
        askpass_script = "askpass_pyautogit"
    #askpass_script_path = os.path.join(askpass_dir, askpass_script)
    environ['GIT_ASKPASS'] = askpass_script
    environ['GIT_USERNAME'] = credentials[0]
    environ['GIT_PASSWORD'] = credentials[1]
    out, err = handle_basic_command(command, command)

    return out, err


def parse_string_into_executable_command(command, remove_quotes):
    """Function that takes in a string command, and parses it into a subprocess arg list

    Parameters
    ----------
    command : str
        The command as a string

    Returns
    -------
    run_command : list of str
        The command as a list of subprocess args
    """

    if '"' in command:
        run_command = []
        strings = re.findall('"[^"]*"', command)
        non_strings = re.split('"[^"]*"', command)
        for i in range(len(strings)):
            run_command = run_command + non_strings[i].strip().split(' ')
            string_in = strings[i]
            if remove_quotes:
                string_in = string_in[1:]
                string_in = string_in[:(len(string_in) - 1)]
            run_command.append(string_in)
        if len(non_strings) == (len(strings) + 1) and len(non_strings[len(strings)]) > 0:
            run_command.append(non_strings[len(strings) + 1])
    else:
        run_command = command.split(' ')

    return run_command


def handle_basic_command(command, name, remove_quotes=True):
    """Function that executes any git command given, and returns program output.

    Parameters
    ----------
    command : str
        The command string to run
    name : str
        The name of the command being run
    remove_quotes : bool
        Since subprocess takes an array of strings, we split on spaces, however in some cases we want quotes to remain together (ex. commit message)
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    out = None
    err = 0

    run_command = parse_string_into_executable_command(command, remove_quotes)
    try:
        LOGGER.write('Executing command: {}'.format(str(run_command)))
        proc = Popen(run_command, stdout=PIPE, stderr=PIPE)
        output, error = proc.communicate()
        if proc.returncode != 0:
            out = error.decode()
            err = proc.returncode
        else:
            out = output.decode()
    except:
        out = "Unknown error processing function: {}".format(name)
        err = -1
    return out, err


def handle_open_external_program_command(command, name):
    """Function used to run commands that open an external program and detatch from pyautogit.

    Parameters
    ----------
    command : str
        Command string to run
    name : str
        Name of command to run
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    run_command = command.split(' ')
    try:
        LOGGER.write('Opening external program with: {}'.format(str(run_command)))
        proc = Popen(run_command, stdout=PIPE, stderr=PIPE)
        out, err_messg = proc.communicate()
        err = proc.returncode
        if err != 0:
            out = err_messg.decode()
        else:
            out = out.decode()
    except FileNotFoundError:
        out = "Program: {} could not be found in system path".format(command.split(' ')[0])
        err = -1
    except:
        out = "Unknown error processing function: {}".format(name)
        err = -1
    return out, err


def handle_custom_command(command):
    """Function that executes a custom, non-git command

    Patameters
    ----------
    command : str
        Command string to run

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    name = command
    return handle_basic_command(command, name)


def open_default_editor(default_editor, path):
    """Function used to open the selected default editor in external window.

    Parameters
    ----------
    default_editor : str
        Editor open command. ex: emacs, code
    path : str
        The path to the file or directory to open

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = '{} {}'.format(default_editor, path)
    name = "open_default_editor"
    return handle_open_external_program_command(command, name)


#####################################################################
#                                                                   #
#                   Git Command Functions Below                     #
#                                                                   #
#####################################################################

#---------------------#
# Git Status Commands #
#---------------------#

def git_status_short(repo_path='.'):
    """Function for getting shorthand git status

    Parameters
    ----------
    repo_path : str
        Target repo path

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = "git -C {} status -s".format(repo_path)
    name = "git_short_status"
    return handle_basic_command(command, name)


def git_status(repo_path='.'):
    """Function for getting git status

    Parameters
    ----------
    repo_path : str
        Target repo path

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = "git -C {} status".format(repo_path)
    name = "git_short_status"
    return handle_basic_command(command, name)


def git_tree(branch):
    """Function that gets git log as a tree

    Parameters
    ----------
    branch : str
        branch or tag name to log

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git --no-pager log --oneline --decorate --all --graph {}'.format(branch)
    name = 'git_tree'
    return handle_basic_command(command, name)


def git_log(branch):
    """Function that gets git log information

    Parameters
    ----------
    branch : str
        branch or tag name to log

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git --no-pager log {}'.format(branch)
    name='git_log'
    return handle_basic_command(command, name)


def git_diff():
    """Function that gets git diff

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git diff'
    name='git_diff'
    return handle_basic_command(command, name)


def git_diff_file(filename):
    """Function that gets git diff for specific file

    Parameters
    ----------
    filename : str
        Name of file to diff

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git diff {}'.format(filename)
    name='git_diff_file'
    return handle_basic_command(command, name)


#---------------------#
# Git Remote Commands #
#---------------------#

def git_get_remotes():
    """Function for returning git remotes list

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = "git remote"
    name = "git_get_remotes"
    return handle_basic_command(command, name)


def git_get_remote_info(remote):
    """Function that gets information about a remote

    Parameters
    ----------
    remote : str
        Name of target remote

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git remote show -n {}'.format(remote)
    name='git_get_remote_info'
    return handle_basic_command(command, name)


def git_add_remote(remote_name, remote_url):
    """Function that adds a new remote to the repository

    Parameters
    ----------
    remote_name : str
        Name of the new remote
    remote_url : str
        URL of the new remote

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git remote add {} {}'.format(remote_name, remote_url)
    name = 'git_add_remote'
    return handle_basic_command(command, name)


def git_remove_remote(remote_name):
    """Function that removes a remote from the repository

    Parameters
    ----------
    remote_name : str
        Name of the remote

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git remote rm {}'.format(remote_name)
    name = 'git_remove_remote'
    return handle_basic_command(command, name)


def git_rename_remote(remote, new_name):
    """Function that renames a remote in the repository

    Parameters
    ----------
    remote : str
        Old name of the remote
    new_name : str
        New name of the new remote

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git remote rename {} {}'.format(remote, new_name)
    name = 'git_rename_remote'
    return handle_basic_command(command, name)


#---------------------#
# Git Commit Commands #
#---------------------#

def git_get_commit_info(commit_hash):
    """Function that gets info about a particular commit.

    Parameters
    ----------
    commit_hash : str
        Hash code for target commit

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git show {}'.format(commit_hash)
    name = 'git_get_commit_info'
    return handle_basic_command(command, name)


def git_checkout_commit(commit_hash):
    """Function that checks out a particular commit.

    Parameters
    ----------
    commit_hash : str
        Hash code for target commit

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git checkout {}'.format(commit_hash)
    name = 'git_checkout_commit'
    return handle_basic_command(command, name)


def git_commit_changes(commit_message):
    """Function that commits added changes

    Parameters
    ----------
    commit_message : str
        Message attached to target commit

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    if len(commit_message) == 0:
        return "No commit message entered", -1
    else:
        command = 'git commit -m "{}"'.format(commit_message)
        name = "git_commit_changes"
        return handle_basic_command(command, name)


def git_create_tag(tag_name):
    """Function that creates a new tag

    Parameters
    ----------
    tag_name : str
        The name of the new tag
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git tag {}'.format(tag_name)
    name = 'git_create_tag'
    return handle_basic_command(command, name)


def git_get_tags():
    """Function that gets list of git tags in repo
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git tag'
    name = 'git_get_tags'
    return handle_basic_command(command, name)

#---------------------#
# Git Branch Commands #
#---------------------#

def git_get_branches():
    """Function that gets a list of the repo branches.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = "git branch"
    name = "git_get_branches"
    return handle_basic_command(command, name)


def git_get_recent_commits(branch):
    """Gets recent commits made to the branch

    Parameters
    ----------
    branch : str
        Name of current branch

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git --no-pager log {} --oneline'.format(branch)
    name = "git_get_recent_commits"
    return handle_basic_command(command, name)


def git_create_new_branch(branch, checkout=True):
    """Creates anew branch for the repo

    Parameters
    ----------
    branch : str
        Name of new branch
    checkout : bool
        If true, checkout branch after creation.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git checkout -b {}'.format(branch)
    name = 'git_create_new_branch'
    return handle_basic_command(command, name)


def git_delete_branch(branch):
    """Deletes existing git branch

    Parameters
    ----------
    branch : str
        Name of branch to delete
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """
    command = 'git branch -d {}'.format(branch)
    name = 'git_delete_branch'
    return handle_basic_command(command, name)


def git_checkout_branch(branch):
    """Checks out given branch

    Parameters
    ----------
    branch : str
        Name of target branch

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git checkout {}'.format(branch)
    name = 'git_checkout_branch'
    return handle_basic_command(command, name)


def git_checkout_tag(tag):
    """Checks out given tag

    Parameters
    ----------
    tag : str
        Name of tag to check out
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git checkout -q {}'.format(tag)
    name = 'git_checkout_tag'
    return handle_basic_command(command, name)


def git_merge_branches(merge_branch):
    """Merges checked out branch with given branch

    Parameters
    ----------
    merge_branch : str
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git merge {}'.format(merge_branch)
    name = 'git_merge_branches'
    return handle_basic_command(command, name)


def git_revert_branch_merge():
    """Undos merge between two branches
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git reset --hard ORIG_HEAD'
    name = 'git_revert_branch_merge'
    return handle_basic_command(command, name)

#--------- ---------#
# Git Repo Commands #
#-------------------#

def git_init_new_repo(new_dir_target):
    """Function that creates a new git repository

    Parameters
    ----------
    new_dir_target : str
        Name of new repo
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    if os.path.exists(new_dir_target):
        err = -1
        out = "Path already exists"
    else:
        os.mkdir(new_dir_target)
        readme_fp = open(os.path.join(new_dir_target, 'README.md'), 'w')
        readme_fp.write('# {}'.format(new_dir_target))
        readme_fp.close()
        command = 'git init {}'.format(new_dir_target)
        name = 'git_init_new_repo'
        out, err = handle_basic_command(command, name)

    return out, err


def git_clone_new_repo(new_repo_url, credentials):
    """Function that clones a new git repository

    Parameters
    ----------
    new_repo_url : str
        URL of new repo
    credentials : list of str
        Username and Password for git remote
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    out = None
    err = 0
    if os.path.exists(new_repo_url.split('/')[-1]):
        err = -1
        out = "The target repo couldn't be cloned - Directory exists"
    else:
        command = 'git clone {}'.format(new_repo_url)
        out, err = handle_credential_command(command, credentials)
        if err == 0:
            out = "Successfully cloned {}".format(new_repo_url)
            
    return out, err

#------------------------#
# Git (Un)Stage Commands #
#------------------------#

def git_add_all():
    """Function that stages all files in repo for commit.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git add -A'
    name = 'git_add_all'
    return handle_basic_command(command, name)


def git_reset_all():
    """Function that unstages all files in repo for commit.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git reset HEAD'
    name = 'git_reset_all'
    return handle_basic_command(command, name)


def git_add_file(filename):
    """Function that stages single file in repo for commit.

    Parameters
    ----------
    filename : str
        Name of file to stage

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git add {}'.format(filename)
    name = 'git_add_file'
    return handle_basic_command(command, name)


def git_reset_file(filename):
    """Function that unstages single file in repo for commit.

    Parameters
    ----------
    filename : str
        Name of file to unstage

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git reset HEAD {}'.format(filename)
    name = 'git_reset_file'
    return handle_basic_command(command, name)


#--------------------#
# Git Stash Commands #
#--------------------#


def git_stash_all():
    """Function that stashes all changes in repo.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git stash'
    name = 'git_stash_all'
    return handle_basic_command(command, name)

def git_unstash_all():
    """Function that unstashes all changes in repo.

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """
    command = 'git stash pop'
    name = 'git_unstash_all'
    return handle_basic_command(command, name)


def git_stash_file(filename):
    """Function that stashes single file in repo.

    Parameters
    ----------
    filename : str
        Name of file to stash

    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git stash {}'.format(filename)
    name = 'git_stash_file'
    return handle_basic_command(command, name)


#------------------------#
# Git Push/Pull Commands #
#------------------------#

def git_pull_branch(branch, remote, credentials):
    """Function that pulls a branch from the remote repo
    
    Parameters
    ----------
    branch : str
        Name of current branch
    remote : str
        Name of remote
    credentials : list of str
        Username and Password of user for remoe
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git pull {} {}'.format(remote, branch)
    return handle_credential_command(command, credentials)


def git_push_to_branch(branch, remote, credentials, repo_path='.'):
    """Function that pushes a branch to the remote repo
    
    Parameters
    ----------
    branch : str
        Name of current branch
    remote : str
        Name of remote
    credentials : list of str
        Username and Password of user for remote
    repo_path : str
        The repository path
    
    Returns
    -------
    out : str
        Output string from stdout if success, stderr if failure
    err : int
        Error code if failure, 0 otherwise.
    """

    command = 'git push {} {}'.format(remote, branch)
    return handle_credential_command(command, credentials)



#!/usr/bin/env python3

from sys import argv
from os import environ

def main():

    if 'username' in argv[1].lower():
        print(environ['GIT_USERNAME'])
        exit()

    if 'password' in argv[1].lower():
        print(environ['GIT_PASSWORD'])
        exit()

    exit(1)


"""Pyautogit internal editor, based on snano example from py_cui
"""

import os
import shutil
import py_cui.widget_set
import pyautogit.screen_manager

class EditorScreenManager(pyautogit.screen_manager.ScreenManager):
    """Class representing internal editor screen for pyautogit

    Attributes
    ----------
    opened_path : str
        The current opened path for the editor
    is_new_file_open : bool
        Flag that says if a new file is open
    """

    def __init__(self, top_manager, opened_path):
        """Contructor for the EditorScreenManager
        """

        super().__init__(top_manager, 'editor')
        if os.path.isdir(opened_path):
            self.opened_path = opened_path
        else:
            self.opened_path = os.dirname(opened_path)

        self.new_file_open = False


    def initialize_screen_elements(self):
        """Override of base class. Initializes editor widgets and widget set

        Returns
        -------
        pyautogit_editor_widget_set : py_cui.widget_set.WidgetSet
            Widget set for internal editor screen
        """

        pyautogit_editor_widget_set = self.manager.root.create_new_widget_set(7, 8)

        self.current_dir_textbox = pyautogit_editor_widget_set.add_text_box('Current Directory', 6, 0, column_span=8)
        self.current_dir_textbox.add_key_command(py_cui.keys.KEY_ENTER, self.open_new_directory)
        self.current_dir_textbox.set_text(self.opened_path)
        self.current_dir_textbox.set_focus_text('Open Directory - Enter | Cancel - Esc')

        self.edit_text_block = pyautogit_editor_widget_set.add_text_block('Open file', 0, 2, row_span=6, column_span=6)
        self.edit_text_block.set_focus_text('Save - Esc + S | Return - Esc')

        self.new_dir_textbox = pyautogit_editor_widget_set.add_text_box('Add New Directory', 4, 0, column_span=2)
        self.new_dir_textbox.add_key_command(py_cui.keys.KEY_ENTER, self.add_new_directory)
        self.new_dir_textbox.set_focus_text('Create Directory - Enter | Cancel - Esc')

        self.new_file_textbox = pyautogit_editor_widget_set.add_text_box('Add New File', 5, 0, column_span=2)
        self.new_file_textbox.add_key_command(py_cui.keys.KEY_ENTER, self.add_new_file)
        self.new_file_textbox.set_focus_text('Create File - Enter | Cancel - Esc')

        self.file_menu = pyautogit_editor_widget_set.add_scroll_menu('Directory Files', 0, 0, row_span=4, column_span=2)
        self.file_menu.add_key_command(py_cui.keys.KEY_ENTER,   self.open_file_dir)
        self.file_menu.add_key_command(py_cui.keys.KEY_DELETE,  self.delete_selected_file_dir)
        self.file_menu.add_key_command(py_cui.keys.KEY_F_LOWER, lambda : self.manager.root.move_focus(self.new_file_textbox))
        self.file_menu.add_key_command(py_cui.keys.KEY_D_LOWER, lambda : self.manager.root.move_focus(self.new_dir_textbox))
        self.file_menu.add_key_command(py_cui.keys.KEY_R_LOWER, self.refresh_status)
        self.file_menu.add_key_command(py_cui.keys.KEY_BACKSPACE, self.manager.open_autogit_window_target)
        self.file_menu.add_text_color_rule('<DIR>', py_cui.GREEN_ON_BLACK, 'startswith', match_type='region', region=[5,1000])
        self.file_menu.set_focus_text('Return -Bcksp | Open - Enter | New File - f | New Dir - d | Refresh - r | Delete - Del | Return - Esc')

        self.info_panel = self.edit_text_block

        pyautogit_editor_widget_set.add_key_command(py_cui.keys.KEY_S_LOWER, self.save_opened_file)
        pyautogit_editor_widget_set.add_key_command(py_cui.keys.KEY_M_LOWER, lambda : self.manager.root.move_focus(self.file_menu))
        pyautogit_editor_widget_set.add_key_command(py_cui.keys.KEY_BACKSPACE, self.manager.open_autogit_window_target)

        self.open_new_directory()

        return pyautogit_editor_widget_set


    def refresh_status(self):
        """Function that refreshes the view of the file menu on new dir or file creation
        """

        self.open_new_directory_external(self.current_dir_textbox.get())


    def set_initial_values(self):
        """Function that sets status bar text
        """

        self.manager.root.set_status_bar_text('Return - Bcksp | Save Open File - s | Open File Menu - m | Navigate - Arrows')


    def clear_elements(self):
        """Function for clearing widgets in editor screen
        """

        self.info_panel.clear()
        self.info_panel.title = 'Open file'
        self.current_dir_textbox.clear()
        self.new_file_textbox.clear()
        self.file_menu.clear()


    def open_new_directory_external(self, new_dir_path):
        """Opens a new directory given an external target

        Parameters
        ----------
        new_dir_path : str
            Path of directory to open
        """

        self.current_dir_textbox.set_text(new_dir_path)
        self.open_new_directory()


    def open_new_directory(self):
        """Function that opens a new directory
        """

        target = self.current_dir_textbox.get()
        if len(target) == 0:
            target = '.'
        elif not os.path.exists(target):
            self.manager.root.show_error_popup('Does not exist', 'ERROR - {} path does not exist'.format(target))
            return
        elif not os.path.isdir(target):
            self.manager.root.show_error_popup('Not a Dir', 'ERROR - {} is not a directory'.format(target))
            return
        target = os.path.abspath(target)
        self.current_dir_textbox.set_text(target)
        self.opened_path = target

        files = []
        files.append('<DIR> ..')
        dir_contents = os.listdir(self.opened_path)
        for elem in dir_contents:
            if os.path.isfile(os.path.join(self.opened_path, elem)):
                files.append(elem)
            else:
                files.append('<DIR> ' + elem)

        self.file_menu.clear()
        self.file_menu.add_item_list(files)
        

    def add_new_file(self):
        """Function for creating a new file
        """

        if os.path.exists(os.path.join(self.current_dir_textbox.get(), self.new_file_textbox.get())):
            self.manager.root.show_error_popup('File Exists', 'File exists with name {}!'.format(self.new_file_textbox.get()))
        else:
            self.edit_text_block.title = 'Open file - ' + self.new_file_textbox.get() + ' - (Unsaved)'
            self.edit_text_block.clear()
            self.new_file_textbox.clear()
            self.manager.root.move_focus(self.edit_text_block)


    def add_new_directory(self):
        """Function for creating a new directory
        """

        try:
            os.mkdir(self.new_dir_textbox.get())
            self.new_dir_textbox.clear()
            self.refresh_status()
            self.manager.root.move_focus(self.file_menu)
        except OSError:
            self.manager.root.show_error_popup('OS ERROR', 'Failed to create new directory!')


    def open_file_dir(self):
        """Function that opens a file/directory from menu
        """

        filename = self.file_menu.get()
        if filename.startswith('<DIR>'):
            self.current_dir_textbox.set_text(os.path.join(self.opened_path, filename[6:]))
            self.open_new_directory()
        else:
            try:
                self.edit_text_block.text_color_rules = []
                fp = open(os.path.join(self.opened_path, filename), 'r')
                text = fp.read()
                fp.close()
                self.edit_text_block.set_text(text)
                self.edit_text_block.title = 'Open file - {}'.format(filename)
            except:
                self.manager.root.show_warning_popup('Not a text file', 'The selected file could not be opened - not a text file')


    def save_opened_file(self):
        """Function that saves the opened file
        """

        if self.edit_text_block.title.startswith('Open file - '):
            filename = self.edit_text_block.title.split('-', 1)[1].strip()
            fp = open(os.path.join(self.opened_path, filename), 'w')
            fp.write(self.edit_text_block.get())
            fp.close()
            self.refresh_status()
            self.manager.root.move_focus(self.edit_text_block)
            self.edit_text_block.title = 'Open file - {}'.format(filename)
        else:
            self.manager.root.show_error_popup('No File Opened', 'Please open a file before saving it.')


    def delete_selected_file_dir(self):
        """Function that deletes the selected file
        """

        if not self.file_menu.get().startswith('<DIR>'):
            target = os.path.join(self.current_dir_textbox.get(), self.file_menu.get())
            try:
                os.remove(target)
            except:
                self.manager.root.show_error_popup('Delete Error', 'Failed to delete the target file!')
        else:
            target = os.path.join(self.current_dir_textbox.get(), self.file_menu.get()[6:])
            try:
                shutil.rmtree(target)
            except:
                self.manager.root.show_error_popup('Delete Error', 'Could not remove target directory!')

        self.refresh_status()




"""Module containing logging classes and functions.

The logger is controlled via a set of global variables set by the pyautogit client.
"""

import os
import datetime

# Global var that stores path to logfile
_LOG_FILE_PATH = None

# Global var that stores whether or not logging is enabled
_LOG_ENABLED = False

# Global var that stores file pointer for log file
_LOG_FILE_POINTER = None


def toggle_logging():
    """Function for opening/closing log file as required.
    """

    global _LOG_ENABLED
    global _LOG_FILE_PATH
    if _LOG_FILE_PATH is None and not _LOG_ENABLED:
        return
    if _LOG_ENABLED:
        close_logger()
        _LOG_ENABLED = False
    elif not os.path.exists(os.path.dirname(_LOG_FILE_PATH)):
        return
    else:
        if _LOG_FILE_PATH is not None and os.access(os.path.dirname(_LOG_FILE_PATH), os.W_OK):
            initialized = initialize_logger()
            _LOG_ENABLED = initialized
        else:
            return
                

def set_log_file_path(log_file_path):
    """Sets the path to the log file

    Parameters
    ----------
    log_file_path : str
        Path to the log file
    """

    global _LOG_FILE_PATH
    _LOG_FILE_PATH = log_file_path


def initialize_logger():
    """Function for initializing log-file writing in addition to stdout output

    Returns
    -------
    initialized : bool
        True if log file opened, false otherwise
    """

    global _LOG_FILE_PATH
    global _LOG_FILE_POINTER
    if os.path.exists(_LOG_FILE_PATH):
        if not os.access(_LOG_FILE_PATH, os.W_OK):
            return False
    _LOG_FILE_POINTER = open(_LOG_FILE_PATH, 'w+')
    return True


def close_logger():
    """Function that closes the opened logfile
    """

    global _LOG_FILE_POINTER
    if _LOG_FILE_POINTER is not None:
        _LOG_FILE_POINTER.close()
        _LOG_FILE_POINTER = None



def write(text, no_timestamp=False):
    """Main logging funcion. Called if write function was set
    
    Parameters
    ----------
    text : str
        debug text to print
    no_timestamp=False : bool
        a flag to disable timestamp printing when required
    """

    global _LOG_FILE_POINTER
    global _LOG_ENABLED
    if _LOG_ENABLED and _LOG_FILE_POINTER is not None:

        # remove timestamp if not in use
        if no_timestamp:
            final_text = '{}\n'.format(text)
        else:
            # otherwise add timestamp
            final_text = '{} - {}\n'.format(datetime.datetime.now(), text)

        _LOG_FILE_POINTER.write(final_text)




"""File containing functions used by the repository specific CUI screen.

This file is meant to handle the intermediate considerations between the 
CUI and the underlying git commands found in pyautogit.commands

Author: Jakub Wlodek  
Created: 01-Oct-2019
"""

import os
from sys import platform
import py_cui
import pyautogit
import pyautogit.commands
import pyautogit.screen_manager
import pyautogit.logger as LOGGER

class RepoControlManager(pyautogit.screen_manager.ScreenManager):
    """Class responsible for managing functions for the repository control screen.

    This class contains functions that are used by pyautogit for individual repository control.
    It provides the interface between the CUI widgets for the repository control screen and
    the pyautogit.commands module.

    Attributes
    ----------
    menu_choices : list of str
        Overriden list of menu choices accessible from the repository control menu
    """

    def __init__(self, top_manager):
        """Constructor for the RepoControlManager class
        """

        super().__init__(top_manager, 'repository control')
        self.menu_choices = ['(Re)Enter Credentials', 
                                'Push Branch', 
                                'Pull Branch', 
                                'Add Remote', 
                                'Add All', 
                                'Stash All', 
                                'Stash Pop',
                                'Open Repository in Editor', 
                                'Enter Custom Command', 
                                'About',
                                'Exit']


    def process_menu_selection(self, selection):
        """Override of base class, executes based on user menu selection

        Parameters
        ----------
        selection : str
            User selection from menu
        """

        if selection == 'Add Remote':
            self.ask_new_remote_name()
        elif selection == 'Add All':
            self.add_all_changes()
        elif selection == 'Push Branch':
            self.execute_long_operation('Pushing', self.push_repo_branch, credentials_required=True)
        elif selection == 'Pull Branch':
            self.execute_long_operation('Pulling', self.pull_repo_branch, credentials_required=True)
        elif selection == 'Stash All':
            self.execute_long_operation('Stashing', self.stash_all_changes, credentials_required=False)
        elif selection == 'Stash Pop':
            self.execute_long_operation('Unstashing', self.unstash_all_changes, credentials_required=False)
        elif selection == 'Checkout Version':
            self.show_version_selection_screen()
        elif selection == 'About':
            self.info_panel.set_text(self.manager.get_about_info())
        elif selection == 'Open Repository in Editor':
            self.open_editor()
        elif selection == '(Re)Enter Credentials':
            self.manager.ask_credentials()
        elif selection == 'Enter Custom Command':
            self.ask_custom_command()
        elif selection == 'Exit':
            self.manager.close_cleanup()
            exit()
        else:
            self.manager.open_not_supported_popup(selection)


    def initialize_screen_elements(self):
        """Function that initializes the widgets for the repo control screen. Override of base class function

        Returns
        -------
        repo_control_widget_set : py_cui.widget_set.WidgetSet
            Widget set object for repo control screen
        """

        # Repo Control window screen widgets, key commands.
        repo_control_widget_set = self.manager.root.create_new_widget_set(9, 8)

        # Base keyboard shortcuts.
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_BACKSPACE, self.manager.open_repo_select_window)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_R_LOWER, self.refresh_status)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_M_LOWER, self.show_menu)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_L_LOWER, self.show_log)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_A_LOWER, self.add_all_changes)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_E_LOWER, self.open_editor)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_F_LOWER, lambda : self.execute_long_operation('Pulling', self.pull_repo_branch, credentials_required=True))
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_P_LOWER, lambda : self.execute_long_operation('Pushing', self.push_repo_branch, credentials_required=True))
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_H_LOWER, self.show_help_overview)
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_C_UPPER, self.ask_custom_command)

        # Textboxes for commit message and new branch/tag
        self.new_branch_textbox = repo_control_widget_set.add_text_box('New Branch', 8, 0, column_span=2)
        self.new_branch_textbox.add_key_command(py_cui.keys.KEY_ENTER, self.create_new_branch)
        self.new_branch_textbox.set_focus_text('Create New Branch - Enter | Return - Esc')
        
        self.commit_message_box = repo_control_widget_set.add_text_box('Commit Message', 8, 2, column_span=6)
        self.commit_message_box.add_key_command(py_cui.keys.KEY_ENTER, self.commit)
        self.commit_message_box.set_focus_text('Commit Changes - Enter | Return - Esc')

        # Add/Unstage files menu. Shows current git status
        self.add_files_menu = repo_control_widget_set.add_scroll_menu('Add Files', 0, 0, row_span=2, column_span=2)
        self.add_files_menu.add_text_color_rule(' ', py_cui.RED_ON_BLACK,   'startswith',       match_type='region', region=[0,3], include_whitespace=True)
        self.add_files_menu.add_text_color_rule('?', py_cui.RED_ON_BLACK,   'startswith',       match_type='region', region=[0,3], include_whitespace=True)
        self.add_files_menu.add_text_color_rule(' ', py_cui.GREEN_ON_BLACK, 'notstartswith',    match_type='region', region=[0,3], include_whitespace=True)
        self.add_files_menu.add_key_command(py_cui.keys.KEY_ENTER,      self.add_revert_file)
        self.add_files_menu.add_key_command(py_cui.keys.KEY_SPACE,      self.open_git_diff_file)
        self.add_files_menu.add_key_command(py_cui.keys.KEY_E_LOWER,    self.open_editor_file)
        self.add_files_menu.add_key_command(py_cui.keys.KEY_H_LOWER,    self.show_help_add_files_menu)
        self.add_files_menu.set_focus_text('Add/Unstage - Enter | Diff - Space | Edit - e | Help - h | Return - Esc')

        # Shows current git remotes
        self.remotes_menu = repo_control_widget_set.add_scroll_menu('Git Remotes', 2, 0, row_span=2, column_span=2)
        self.remotes_menu.add_key_command(py_cui.keys.KEY_ENTER,    self.show_remote_info)
        self.remotes_menu.add_key_command(py_cui.keys.KEY_N_LOWER,  self.ask_new_remote_name)
        self.remotes_menu.add_key_command(py_cui.keys.KEY_DELETE,   self.delete_remote)
        self.remotes_menu.add_key_command(py_cui.keys.KEY_R_LOWER,  lambda : self.manager.ask_message('Please enter a new remote name', callback=self.rename_remote))
        self.remotes_menu.add_key_command(py_cui.keys.KEY_H_LOWER,  self.show_help_remotes_menu)
        self.remotes_menu.set_focus_text('Remote Info - Enter | New Remote - n | Delete Remote - Delete | Help - h | Return - Esc')

        # Shows current git branches/tags
        self.branch_menu_state = 'branches'
        self.branch_menu = repo_control_widget_set.add_scroll_menu('Git Branches', 4, 0, row_span=2, column_span=2)
        self.branch_menu.add_key_command(py_cui.keys.KEY_ENTER,     self.checkout_branch)
        self.branch_menu.add_key_command(py_cui.keys.KEY_SPACE,     self.show_log)
        self.branch_menu.add_key_command(py_cui.keys.KEY_TAB,       self.show_tree)
        self.branch_menu.add_key_command(py_cui.keys.KEY_N_LOWER,   lambda : self.manager.root.move_focus(self.new_branch_textbox))
        self.branch_menu.add_key_command(py_cui.keys.KEY_T_LOWER,   self.show_tags)
        self.branch_menu.add_key_command(py_cui.keys.KEY_B_LOWER,   self.show_branches)
        self.branch_menu.add_key_command(py_cui.keys.KEY_M_LOWER,   self.merge_branches)
        self.branch_menu.add_key_command(py_cui.keys.KEY_U_LOWER,   self.revert_merge)
        self.branch_menu.add_key_command(py_cui.keys.KEY_H_LOWER,   self.show_help_branch_menu)
        self.branch_menu.add_key_command(py_cui.keys.KEY_DELETE,    self.delete_branch)
        self.branch_menu.set_focus_text('Checkout - Enter | Log - Space | New - n | Merge - m | Show Tags - t | Show Branches - b | Revert Merge - u | Help - h | Esc - Return')

        # Shows list of recent git commits for checked out branch.
        self.commits_menu = repo_control_widget_set.add_scroll_menu('Recent Commits', 6, 0, row_span=2, column_span=2)
        self.commits_menu.add_key_command(py_cui.keys.KEY_ENTER,        self.show_commit_info)
        self.commits_menu.add_key_command(py_cui.keys.KEY_SPACE,        self.checkout_commit)
        self.commits_menu.add_key_command(py_cui.keys.KEY_H_LOWER,      self.show_help_commits_menu)
        self.commits_menu.add_text_color_rule('^.*? ', py_cui.GREEN_ON_BLACK, 'contains', match_type='regex', include_whitespace=True)
        self.commits_menu.set_focus_text('Commit Info - Enter | Checkout - Space | Help - h | Return - Esc')

        # Main info text block listing information for all pyautogit operations
        self.info_text_block = repo_control_widget_set.add_text_block('Git Info', 0, 2, row_span=8, column_span=6)
        self.info_text_block.add_text_color_rule('+',           py_cui.GREEN_ON_BLACK,  'startswith')
        self.info_text_block.add_text_color_rule('-',           py_cui.RED_ON_BLACK,    'startswith')
        self.info_text_block.add_text_color_rule('commit',      py_cui.YELLOW_ON_BLACK, 'startswith')
        self.info_text_block.add_text_color_rule('Copyright',   py_cui.CYAN_ON_BLACK,   'startswith')
        self.info_text_block.add_text_color_rule('@.*@',        py_cui.CYAN_ON_BLACK,   'contains', match_type='regex')
        self.info_text_block.add_text_color_rule('**    ',      py_cui.RED_ON_BLACK,    'startswith')
        self.info_text_block.add_text_color_rule('\* *\w+ ',    py_cui.CYAN_ON_BLACK,   'contains', match_type='regex')
        #self.info_text_block.selectable = False

        # Add some simple shortcut commands
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_C_LOWER, lambda : self.manager.root.move_focus(self.commit_message_box))
        repo_control_widget_set.add_key_command(py_cui.keys.KEY_I_LOWER, lambda : self.info_text_block.set_text(self.manager.get_about_info()))

        self.info_panel = self.info_text_block
        return repo_control_widget_set


    def clear_elements(self):
        """Function that clears elements of repo control window
        """

        self.commit_message_box.clear()
        self.info_text_block.clear()
        self.info_text_block.title = 'Git Info'
        self.branch_menu.clear()
        self.commits_menu.clear()
        self.add_files_menu.clear()
        self.remotes_menu.clear()
        self.new_branch_textbox.clear()


    def set_initial_values(self):
        """Function that initializes status bar and info text for repo control window
        """

        self.info_text_block.set_text(self.manager.get_about_info())
        self.branch_menu_state = 'branches'
        self.manager.root.set_status_bar_text('Return - Bcksp | Menu - m | Refresh - r | Add All - a | Commit - c | Log - l | Editor - e | Pull - f | Push - p | Help -h')


    def refresh_status(self):
        """Function that refreshes a git repository status
        """

        remote          = self.remotes_menu.get_selected_item_index()
        selected_file   = self.add_files_menu.get_selected_item_index()

        if self.branch_menu_state == 'branches':
            self.get_repo_branches()
            self.branch_menu.title = 'Git Branches'
            self.new_branch_textbox.title = 'New Branch'
            self.new_branch_textbox.update_key_command(py_cui.keys.KEY_ENTER, self.create_new_branch)
            self.new_branch_textbox.set_focus_text('Enter - Create new branch | Esc - Return')
        else:
            self.get_repo_tags()
            self.branch_menu.title = 'Git Tags'
            self.new_branch_textbox.title = 'New Tag'
            self.new_branch_textbox.update_key_command(py_cui.keys.KEY_ENTER, self.create_new_tag)
            self.new_branch_textbox.set_focus_text('Enter - Create new tag | Esc - Return')
        
        self.get_repo_remotes()
        self.get_repo_status_short()
        self.get_recent_commits()

        if len(self.remotes_menu.get_item_list()) > remote:
            self.remotes_menu.set_selected_item_index(remote)
        if len(self.add_files_menu.get_item_list()) > selected_file:
            self.add_files_menu.set_selected_item_index(selected_file)


    def get_repo_status_short(self):
        """Gets shorthand repository status
        """

        out, err = pyautogit.commands.git_status_short()
        self.show_command_result(out, err, show_on_success=False, command_name="Show Status", error_message="Failed to get status")
        self.add_files_menu.clear()
        self.add_files_menu.add_item_list(out.splitlines())


    def get_repo_remotes(self):
        """Gets list of repository remotes
        """

        out, err = pyautogit.commands.git_get_remotes()
        self.show_command_result(out, err, show_on_success=False, command_name="List Remotes", error_message='Cannot get git remotes')
        self.remotes_menu.clear()
        self.remotes_menu.add_item_list(out.splitlines())


    def show_branches(self):
        """Function that swaps to showing branches
        """

        self.branch_menu_state = 'branches'
        self.refresh_status()


    def get_repo_branches(self):
        """Gets list of repository branches
        """

        out, err = pyautogit.commands.git_get_branches()
        self.show_command_result(out, err, show_on_success=False, command_name="List Branches", error_message='Cannot get git branches')
        self.branch_menu.clear()
        self.branch_menu.add_item_list(out.splitlines())
        selected_branch = 0
        for branch in self.branch_menu.get_item_list():
            if branch.startswith('*'):
                break
            selected_branch = selected_branch + 1
        self.branch_menu.set_selected_item_index(selected_branch)


    def show_tags(self):
        """Function that swaps to showing tags
        """
        
        self.branch_menu_state = 'tags'
        self.refresh_status()


    def get_repo_tags(self):
        """Gets list of repository tags
        """

        out, err = pyautogit.commands.git_get_tags()
        self.show_command_result(out, err, show_on_success=False, command_name="List Tags", error_message="Cannot list git tags")
        self.branch_menu.clear()
        tags = out.splitlines()
        tags.reverse()
        self.branch_menu.add_item_list(tags)


    def show_remote_info(self):
        """Gets info about remote
        """

        if self.branch_menu.get() is None:
            return
        remote = self.remotes_menu.get()
        out, err = pyautogit.commands.git_get_remote_info(remote)
        if err < 0:
            self.manager.root.show_error_popup('Cannot get remote info', out)
        else:
            self.info_text_block.clear()
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('{} remote info'.format(remote))


    def show_commit_info(self):
        """Gets info about a particular commit
        """

        if self.commits_menu.get() is None:
            return
        commit_hash = self.commits_menu.get().split(' ', 1)[0]
        out, err = pyautogit.commands.git_get_commit_info(commit_hash)
        if err != 0:
            self.manager.root.show_error_popup('Failed to generate commit info', out)
        else:
            self.info_text_block.clear()
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('Commit info for {}'.format(commit_hash))


    def get_recent_commits(self):
        """Gets list of recent commits to branch
        """

        if self.branch_menu.get() is None:
            return
        if self.branch_menu_state == 'branches':
            branch = self.branch_menu.get()[2:]
            if branch.startswith('(HEAD'):
                branch = branch.split(' ')[-1][:-1]
        else:
            branch = self.branch_menu.get()
        out, err = pyautogit.commands.git_get_recent_commits(branch)
        if err < 0:
            self.root.show_error_popup('Cannot get recent commits', out)
        else:
            self.commits_menu.clear()
            self.commits_menu.add_item_list(out.splitlines())


    def create_new_tag(self):
        """Creates a new tag
        """

        new_tag_name = self.new_branch_textbox.get()
        if new_tag_name is not None and len(new_tag_name) > 0:
            out, err = pyautogit.commands.git_create_tag(new_tag_name)
            self.show_command_result(out, err, command_name='Create Tag', success_message='Tag {} Created'.format(new_tag_name), error_message='Failed to Create Tag')
            self.show_log()
            self.show_tags()
            self.refresh_status()
        else:
            self.manager.show_warning_popup('Warning', 'Cannot create tag with an empty name!')


    def show_log(self):
        """Displays the git log
        """

        if self.branch_menu.get() is None:
            return
        if self.branch_menu_state == 'branches':
            branch = self.branch_menu.get()[2:]
            if branch.startswith('(HEAD'):
                branch = branch.split(' ')[-1][:-1]
        else:
            branch = self.branch_menu.get()
        out, err = pyautogit.commands.git_log(branch)
        if err < 0:
            self.root.show_error_popup('Unable to show git log for branch {}.'.format(branch), out)
        else:
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('Git log')


    def show_tree(self):
        """Displays git log as a tree
        """

        if self.branch_menu.get() is None:
            return
        if self.branch_menu_state == 'branches':
            branch = self.branch_menu.get()[2:]
            if branch.startswith('(HEAD'):
                branch = branch.split(' ')[-1][:-1]
        else:
            branch = self.branch_menu.get()
        out, err = pyautogit.commands.git_tree(branch)
        if err < 0:
            self.root.show_error_popup('Unable to show git log for branch {}.'.format(branch), out)
        else:
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('Git tree')


    def stash_all_changes(self):
        """Stashes all repo changes
        """

        self.message, self.status = pyautogit.commands.git_stash_all()
        self.refresh_status()
        self.manager.root.stop_loading_popup()


    def unstash_all_changes(self):
        """Pops the stash
        """

        self.message, self.status = pyautogit.commands.git_unstash_all()
        self.refresh_status()
        self.manager.root.stop_loading_popup()


    def open_git_diff(self):
        """Opens current git diff state
        """

        out, err = pyautogit.commands.git_diff()
        if err < 0:
            self.manager.root.show_error_popup('Unable to show git diff repo.', out)
        else:
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('Git Diff')


    def open_git_diff_file(self):
        """Gets the diff for a selected file
        """

        filename = self.add_files_menu.get()[3:]
        out, err = pyautogit.commands.git_diff_file(filename)
        if err < 0:
            self.manager.root.show_error_popup('Unable to show git diff for file {}.'.format(filename), out)
        else:
            self.info_text_block.set_text(out)
            self.info_text_block.set_title('Git Diff - {}'.format(filename))


    def open_editor(self, file=None):
        """Opens an external editor if selected
        """

        if self.manager.editor_type == 'Internal':
            self.manager.open_editor_window()
        elif self.manager.default_editor is None:
            self.manager.root.show_error_popup('Error', 'No default editor specified.')
        else:
            current_path = os.getcwd()
            if file is not None:
                current_path = os.path.join(current_path, file)
            out, err = pyautogit.commands.open_default_editor(self.manager.default_editor, current_path)
            if err != 0:
                self.manager.root.show_error_popup('Failed to open editor.', out)
            else:
                self.manager.root.show_message_popup('Opened {}'.format(current_path), 'Opened {} editor in external window'.format(self.manager.default_editor))


    def open_editor_file(self):
        """Opens an external editor for a selected file
        """

        filename = self.add_files_menu.get()[3:]
        self.open_editor(file=filename)


    def add_all_changes(self):
        """Adds all changes to staging
        """

        out, err = pyautogit.commands.git_add_all()
        if err != 0:
            self.manager.root.show_error_popup('Git Add Error', out)
        else:
            self.refresh_status()


    def add_revert_file(self):
        """Adds/Reverts single file from staging
        """

        filename = self.add_files_menu.get()
        if filename.startswith(' ') or filename.startswith('?'):
            out, err = pyautogit.commands.git_add_file(filename[3:])
        else:
            out, err = pyautogit.commands.git_reset_file(filename[3:])
        if err < 0:
            self.manager.root.show_error_popup('Cannot add/revert file {}'.format(filename), out)
        else:
            self.refresh_status()


    #-----------------------------------#
    #               REMOTES             #
    #-----------------------------------#


    def ask_new_remote_name(self):
        """Opens text box to enter new remote name
        """

        self.manager.root.show_text_box_popup('Please enter the new remote name.', self.ask_new_remote_url)


    def ask_new_remote_url(self, remote_name):
        """Opens text box to ask new remote url

        Parameters
        ----------
        remote_name : str
            Remote name entered in previous textbox
        """

        self.utility_var = remote_name
        self.manager.root.show_text_box_popup('Please enter the new remote url.', self.add_remote)


    def add_remote(self, remote_url):
        """Adds remote to git repo

        Parameters
        ----------
        remote_url : str
            URL entered by user
        """

        remote_name = self.utility_var
        self.utility_var = None
        out, err = pyautogit.commands.git_add_remote(remote_name, remote_url)
        self.show_command_result(out, err, show_on_success=False, command_name="Add Remote", error_message="Failed to add remote")
        self.remotes_menu.clear()
        self.refresh_status()


    def delete_remote(self):
        """Deletes selected remote from local repo
        """

        remote = self.remotes_menu.get()
        out, err = pyautogit.commands.git_remove_remote(remote)
        self.show_command_result(out, err, show_on_success=False, command_name="Remove Remote", error_message="Failed to remove remote")
        self.remotes_menu.clear()
        self.refresh_status()


    def rename_remote(self):
        """Renames selected remote from local repo
        """

        new_name = self.manager.user_message
        remote = self.remotes_menu.get()
        out, err = pyautogit.commands.git_rename_remote(remote, new_name)
        self.show_command_result(out, err, show_on_success=False, command_name="Rename Remote", error_message="Failed to rename remote")
        self.remotes_menu.clear()
        self.refresh_status()


    def commit(self):
        """Commits currently staged items
        """

        commit_message = self.commit_message_box.get()
        out, err = pyautogit.commands.git_commit_changes(commit_message)
        self.show_command_result('Commit: {}'.format(commit_message), err, command_name='Commit', success_message='Commit Succeeded', error_message='Commit Failed')
        self.refresh_status()
        self.show_commit_info()
        #self.show_log()
        self.commit_message_box.clear()


    def pull_repo_branch(self):
        """Pulls from remote
        """

        branch = self.branch_menu.get()[2:]
        remote = self.remotes_menu.get()
        self.message, self.status = pyautogit.commands.git_pull_branch(branch, remote, self.manager.credentials)
        self.refresh_status()
        self.manager.root.stop_loading_popup()


    def push_repo_branch(self):
        """Pushes to remote
        """

        branch = self.branch_menu.get()[2:]
        remote = self.remotes_menu.get()
        self.message, self.status = pyautogit.commands.git_push_to_branch(branch, remote, self.manager.credentials)
        if self.status == 0:
            self.message = 'Pushed {} to {} successfully.'.format(branch, remote)
        self.refresh_status()
        self.manager.root.stop_loading_popup()


    def create_new_branch(self):
        """Creates new branch for repository
        """

        new_branch_name = self.new_branch_textbox.get()
        if len(new_branch_name) == 0:
            self.manager.root.show_error_popup('ERROR - Illegal branchname', 'Please enter a valid branchname.')
        else:
            out, err = pyautogit.commands.git_create_new_branch(new_branch_name)
            if err != 0:
                self.manager.show_error_popup('Failed to create branch', out)
            self.manager.root.lose_focus()
            self.new_branch_textbox.clear()
            self.refresh_status()


    def delete_branch(self):
        """Deletes selected branch
        """

        branch_name = self.branch_menu.get()
        if branch_name.startswith('* '):
            self.manager.root.show_error_popup('ERROR - Branch Checked Out', 'You cannot delete the currently checked out branch!')
        elif self.branch_menu_state == 'tags':
            self.manager.root.show_error_popup('ERROR - Tag Menu Open', 'Please open branch menu to delete branches.')
        elif branch_name[2:].startswith('(HEAD'):
            self.manager.root.show_warning_popup('Warning', 'Cannot delete detached head!')
            return
        else:
            out, err = pyautogit.commands.git_delete_branch(branch_name[2:])
            self.show_command_result(out, err, command_name='Delete Branch', success_message='Deleted Branch {}'.format(branch_name[2:]), error_message='Failed To Delete Branch')
            self.refresh_status()


    def checkout_branch(self):
        """Checks out specified branch
        """

        branch = self.branch_menu.get()
        if branch.startswith('* '):
            self.manager.root.show_warning_popup('Warning', 'The selected branch is already checked out!')
            return
        if branch is not None and self.branch_menu_state == 'branches':
            branch = branch[2:]
            if branch.startswith('(HEAD'):
                self.manager.root.show_warning_popup('Warning', 'Cannot checkout detached head!')
                return
            out, err = pyautogit.commands.git_checkout_branch(branch)
            self.show_command_result(out, err, command_name='Branch Checkout', success_message='Checked Out Branch {}'.format(branch), error_message='Failed To Checkout Branch')
            self.refresh_status()
        elif branch is not None:
            out, err = pyautogit.commands.git_checkout_tag(branch)
            self.show_command_result(out, err, command_name='Tag Checkout', success_message='Checked Out Tag {}'.format(branch), error_message='Failed To Checkout Tag')
            self.show_branches()
            self.refresh_status()


    def merge_branches(self):
        """Merges selected branch into the currently checked out branch
        """
        
        merge_branch = self.branch_menu.get()
        checkout_branch = None
        for branch in self.branch_menu.get_item_list():
            if branch.startswith('* '):
                checkout_branch = branch
        if merge_branch is None or checkout_branch is None:
            self.manager.root.show_error_popup('No branches', 'The current repo has no branches to merge!')
        elif merge_branch[2:] == checkout_branch[2:]:
            self.manager.root.show_error_popup('Same branch', 'You cannot merge the same branch with itself!')
        else:
            merge_branch = merge_branch[2:]
            out, err = pyautogit.commands.git_merge_branches(merge_branch)
            self.refresh_status()
            self.show_command_result(out, err, command_name='Merging Branches', success_message='Merged With {}'.format(merge_branch), error_message='Failed To Merge Branch')


    def revert_merge(self):
        """Undos the merge that was just performed
        """

        out, err = pyautogit.commands.git_revert_branch_merge()
        self.show_command_result(out, err, command_name='Revert Merge')
        self.refresh_status()


    def checkout_commit(self):
        """Checks out specified commit
        """

        commit = self.commits_menu.get()
        if commit is not None:
            commit_hash = commit.split(' ', 1)[0]
            out, err = pyautogit.commands.git_checkout_commit(commit_hash)
            self.show_command_result(out, err, command_name='Commit Checkout', success_message='Checked Out Commit {}'.format(commit_hash), error_message='Failed To Checkout Commit')
            self.refresh_status()


    #-------------------------------------------------------#
    # HELP MESSAGE PRINT FUNCTIONS                          #
    #-------------------------------------------------------#


    def show_help_overview(self):
        """Function that displays help message for overview mode.
        """

        help_message = '\n'
        help_message = help_message + 'Welcome to the repository control screen.\nYou are currently in overview mode.\n'
        help_message = help_message + '\nIn overview mode, use the arrow keys to move between widgets.\nUse the Enter key to enter a menu or text field, and "m" to open a popup menu.\n'
        help_message = help_message + '\nTo return to the workspace screen, press Backspace.\nTo create a new commit, either use "a" to add all, or enter the add-files\n'
        help_message = help_message + 'menu to select individual files.\nThen, navigate to the commit message box to add a commit message.\n'
        help_message = help_message + 'You can automatically enter the commit message box from overview mode by pressing "c".\n'
        help_message = help_message + '\nIn addition, to refresh the window based on changes made outside of pyautogit, press "r".\n'
        help_message = help_message + '\nTo show help information for a specific submenu, enter it, and press "h".\n'
        self.info_text_block.set_title('Repo Control Overview Help')
        self.info_text_block.set_text(help_message)


    def show_help_add_files_menu(self):
        """Function that displays help message for the add files menu.
        """

        help_message = '\n'
        help_message = help_message + 'Your currently selected menu is the add files menu.\n'
        help_message = help_message + '\nFrom here, use the arrow keys to scroll, Enter to stage and unstage files for commit.\n'
        help_message = help_message + '\nIf you would like to edit a file, press "e".\nThis will open the internal editor or if specified an external one.\n'
        help_message = help_message + '\nPressing the Space button will display git diff information for the selected file, if any.\n'
        help_message = help_message + '\nTo return to overview mode, press Escape.\n'
        self.info_text_block.set_title('Add Files Menu Help')
        self.info_text_block.set_text(help_message)

    
    def show_help_remotes_menu(self):
        """Function that displays help message for the add files menu.
        """

        help_message = '\n'
        help_message = help_message + 'You have selected the remotes menu.\n\nFrom here, press Enter to show remote info, Delete to delete the remote,\n'
        help_message = help_message + '"n" to create a new remote, and Escape to return to overview mode.\n'
        self.info_text_block.set_title('Remotes Menu Help')
        self.info_text_block.set_text(help_message)


    def show_help_branch_menu(self):
        """Function that displays help message for the add files menu.
        """

        help_message = '\n'
        help_message = help_message + 'This is the branch/tags menu. Use this in combination with the below textbox to\n'
        help_message = help_message + 'control and create new branches or tags. To swap to seeing tags, press "t", and press "b"\n'
        help_message = help_message + 'to go back to viewing branches.\n'
        help_message = help_message + '\nTo create a new branch/tag, select the appropriate mode, and enter the name into the textbox.\n'
        help_message = help_message + '\nTo checkout a branch/tag select it and press enter.\nTo show a log or tree for the branch, press Space or Tab.\n'
        help_message = help_message + '\n To merge two branches together, checkout one and select another and press "m".\nThis will merge the selected one into the checked out one.\n'
        help_message = help_message + '\nTo return to overview mode, press Escape.\n'
        self.info_text_block.set_title('Branch/Tag Menu Help')
        self.info_text_block.set_text(help_message)


    def show_help_commits_menu(self):
        """Function that displays help message for the add files menu.
        """

        help_message = '\n'
        help_message = help_message + 'This is the commits menu. You can check out individual commits and show commit info.\n'
        help_message = help_message + '\nTo check out a commit, select it and press Space, to show info, select and press Enter\n'
        help_message = help_message + '\nTo return to overview mode, press Escape.\n'
        self.info_text_block.set_title('Commits Menu Help')
        self.info_text_block.set_text(help_message)



"""A subscreen that allows for setting a variety of pyautogit settings.
"""

import os
import webbrowser
import datetime
import py_cui.widget_set
import pyautogit
import pyautogit.screen_manager
import pyautogit.logger as LOGGER
import urllib.request
import urllib.error


class SettingsScreen(pyautogit.screen_manager.ScreenManager):
    """Class representing settings subscreen for pyautogit

    Attributes
    ----------
    current_info_log : str
        The current settings log text
    show_settings_log : bool
        Toggle for showing settings log
    """


    def __init__(self, top_manager):
        """Constructor for SettingsScreen
        """

        super().__init__(top_manager, 'settings screen')
        self.current_info_log = ''
        self.show_settings_log = True


    def initialize_screen_elements(self):
        """Override of base class function. Initializes widgets, and returns widget set

        Returns
        -------
        settings_widget_set : py_cui.widget_set.WidgetSet
            Widget set object for rsettings screen
        """

        # Output widget set
        settings_widget_set = self.manager.root.create_new_widget_set(9, 6)
        settings_widget_set.add_key_command(py_cui.keys.KEY_BACKSPACE, self.manager.open_repo_select_window)

        # Logo and link labels
        logo_label = settings_widget_set.add_block_label(self.get_settings_ascii_art(), 0, 0, row_span=2, column_span=3, center=True)
        logo_label.set_color(py_cui.RED_ON_BLACK)
        link_label = settings_widget_set.add_label('Settings Screen - pyautogit v{}'.format(pyautogit.__version__), 0, 3, row_span=2, column_span=3)
        link_label.add_text_color_rule('Settings Screen*', py_cui.CYAN_ON_BLACK, 'startswith', match_type='line')

        # Logging settings
        debug_log_label = settings_widget_set.add_label('Debug Logging', 2, 0)
        debug_log_label.toggle_border()
        self.debug_log_toggle = settings_widget_set.add_button('Toggle Logs', 2, 1, command=self.toggle_logging)
        self.debug_enter_path_button = settings_widget_set.add_button('Set Log File', 2, 2, command=self.ask_log_file_path)
        self.debug_log_status_label = settings_widget_set.add_label('OFF - {}'.format(LOGGER._LOG_FILE_PATH), 3, 0, column_span=3)

        # Default editor settings
        editor_label = settings_widget_set.add_label('Default Editor', 4, 0)
        editor_label.toggle_border()
        self.external_editor_toggle = settings_widget_set.add_button('External/Internal',   4, 1, command=self.toggle_editor_type)
        self.external_editor_enter  = settings_widget_set.add_button('Select Editor',       4, 2, command=self.ask_default_editor)
        self.editor_status_label = settings_widget_set.add_label('{} - {}'.format(self.manager.editor_type, self.manager.default_editor), 5, 0, column_span=3)
        #self.editor_status_label.toggle_border()

        # Program info settings
        about_label = settings_widget_set.add_label('About', 6, 0)
        about_label.toggle_border()
        self.fetch_readme_file_button   = settings_widget_set.add_button('README',          6, 1, command=lambda : self.fetch_about_file('README.md'))
        self.fetch_authors_button       = settings_widget_set.add_button('Authors',         6, 2,  command=lambda : self.fetch_about_file('AUTHORS'))
        self.fetch_license_button       = settings_widget_set.add_button('License',         7, 1, command=lambda : self.fetch_about_file('LICENSE'))
        self.revert_settings_log_button = settings_widget_set.add_button('Settings Log',    7, 2, command=self.revert_settings_log)

        # Documentation settings
        docs_label = settings_widget_set.add_label('Docs', 8, 0)
        docs_label.toggle_border()
        self.show_tutorial_button = settings_widget_set.add_button('Tutorial', 8, 1, command=self.show_tutorial)
        self.open_web_docs_button = settings_widget_set.add_button('Online Docs', 8, 2, command=self.open_web_docs)

        # Info panel
        self.settings_info_panel = settings_widget_set.add_text_block('Settings Info Log', 2, 3, row_span=7, column_span=3)
        self.settings_info_panel.is_selectable = False
        self.info_panel = self.settings_info_panel

        if not LOGGER._LOG_ENABLED:
            self.update_log_file_path('.pyautogit/{}.log'.format(str(datetime.datetime.today()).split(' ')[0]), default_path=True)
        return settings_widget_set


    def set_initial_values(self):
        """Function that sets initial status bar text for settings window
        """

        self.manager.root.set_status_bar_text('Backspace - Return | Enter - Press Buttons | Arrows - Navigate')


    def add_to_settings_log(self, text):
        """Function that updates the settings info log panel

        Parameters
        ----------
        text : str
            New log item to write to settings info panel
        """

        self.current_info_log = '{}\n{}'.format(text, self.current_info_log)
        if self.show_settings_log:
            self.settings_info_panel.set_text(self.current_info_log)


    def fetch_about_file(self, file):
        """Function that grabs file from github and displays it in info panel

        Parameters
        ----------
        file : str
            Filename to fetch from github repository
        """

        self.info_panel.clear()
        self.info_panel.is_selectable = True
        self.show_settings_log = False

        try:
            file_txt = ''
            for line in urllib.request.urlopen('https://raw.githubusercontent.com/jwlodek/pyautogit/master/{}'.format(file)):
                file_txt = '{}{}'.format(file_txt, line.decode('utf-8'))
            self.info_panel.set_text(file_txt)
        except urllib.error.HTTPError:
            self.revert_settings_log()
            self.manager.root.show_error_popup('Http Error', 'Unable to fetch {} file - not found!'.format(file))
        except:
            self.revert_settings_log()
            self.manager.root.show_error_popup('Unknown Error', 'Unable to fetch {} file!'.format(file))


    def revert_settings_log(self):
        """Function that resets to showing settings info
        """

        self.info_panel.is_selectable = False
        if not self.show_settings_log:
            self.show_settings_log = True
        self.info_panel.clear()
        self.info_panel.set_text(self.current_info_log)


    def open_web_docs(self):
        """Function tasked with open docs in external browser
        """

        try:
            webbrowser.open('https://jwlodek.github.io/pyautogit-docs')
        except:
            self.manager.root.show_error_popup('Unknown Error', 'Unable to open online documentation!')


    def show_tutorial(self):
        """Function that demonstrates tutorial for using pyautogit
        """

        self.manager.root.show_warning_popup('Unimplemented', 'The Tutorial feature has not yet been implemented.')


    def ask_log_file_path(self):
        """Prompts user to enter log file path
        """

        self.manager.root.show_text_box_popup('Please enter a new log file path', self.update_log_file_path)


    def get_settings_ascii_art(self):
        """Gets ascii art settings logo

        Returns
        -------
        settings_message : str
            Block letter ascii art settings logo
        """

        settings_message =                      ' ____  ____  ____  ____  __  __ _   ___   ____  \n '
        settings_message = settings_message +   '/ ___)(  __)(_  _)(_  _)(  )(  ( \ / __) / ___) \n'
        settings_message = settings_message +   '\___ \ ) _)   )(    )(   )( /    /( (_ \ \___ \ \n'
        settings_message = settings_message +   '(____/(____) (__)  (__) (__)\_)__) \___/ (____/ '
        return settings_message


    def toggle_editor_type(self):
        """Function that toggles between internal and external editor
        """

        if self.manager.editor_type == 'Internal':
            self.manager.editor_type = 'External'
        else:
            self.manager.editor_type = 'Internal'
        self.add_to_settings_log('Swapped editor type')
        self.refresh_status()


    def toggle_logging(self):
        """Function that enables/disables logging
        """
        
        LOGGER.toggle_logging()
        self.add_to_settings_log('Toggled logging')
        self.refresh_status()


    def ask_default_editor(self):
        """Function that asks user for editor, and then refreshes
        """

        self.manager.root.show_text_box_popup('Please enter a command to open an external text editor', self.update_default_editor)


    def update_default_editor(self, new_editor):
        """Function that updates the new default editor

        Parameters
        ----------
        new_editor : str
            command used to open external editor
        """

        self.manager.default_editor = new_editor
        if self.manager.editor_type == 'Internal':
            self.toggle_editor_type()
        self.add_to_settings_log('Update default editor to: {}'.format(new_editor))
        self.refresh_status()


    def update_log_file_path(self, new_log_file_path, default_path=False):
        """Function that updates log file path if valid

        Parameters
        ----------
        new_log_file_path : str
            Path to new log file
        """

        if os.path.exists(os.path.dirname(new_log_file_path)) and os.access(os.path.dirname(new_log_file_path), os.W_OK):
            LOGGER.set_log_file_path(new_log_file_path)
            self.add_to_settings_log('Update log file path: {}'.format(new_log_file_path))
            self.refresh_status()
        elif not default_path:
            self.manager.root.show_error_popup('Permission Error', 'The log file path either does not exist, or you do not have write permissions!')


    def refresh_status(self):
        """Override of base class refresh function.
        """

        logging_on_off = 'OFF'
        if LOGGER._LOG_ENABLED:
            logging_on_off = 'ON'
        log_file_path = LOGGER._LOG_FILE_PATH
        self.debug_log_status_label.set_title('{} - {}'.format(logging_on_off, log_file_path))

        self.editor_status_label.set_title('{} - {}'.format(self.manager.editor_type, self.manager.default_editor))


from setuptools import setup, find_packages
from sys import platform
from setuptools.command.install import install
import os

with open('requirements.txt', 'r') as req_fp:
    required_packages = req_fp.readlines()

# Use README for long description
with open('README.md', 'r') as readme_fp:
    long_description = readme_fp.read()


# This class is required since the askpass helper scripts need to be installed with execute permissions
class InstallLibrary(install):
    def run(self):
        install.run(self)
        for fn in self.get_outputs():
            if 'askpass' in fn and fn.endswith('.py'):
                os.chmod(fn, 0o777)

setup(
    name="pyautogit",
    version="0.0.5",
    author="Jakub Wlodek",
    author_email="jwlodek.dev@gmail.com",
    description="A command line interface for working with git written in python with the help of py_cui.",
    license="BSD 3-Clause",
    keywords="git command-line cli cui curses",
    url="https://github.com/jwlodek/pyautogit",
    long_description=long_description,
    long_description_content_type='text/markdown',
    cmdclass={'install':InstallLibrary},
    packages = find_packages(exclude=['tests', 'docs']),
    extras_require={
        'test': ['pytest'],
    },
    entry_points={
        'console_scripts': [
            'pyautogit = pyautogit:main',
            'askpass_pyautogit = pyautogit.askpass.askpass:main',
            'askpass_pyautogit_win = pyautogit.askpass.askpass_win:main'
        ],
    },
    install_requires=required_packages,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],

)



"""Manager implementation for CUI screen for selecting different repositories.
"""

import py_cui
import pyautogit
import pyautogit.commands
import pyautogit.screen_manager
import pyautogit.logger as LOGGER


class RepoSelectManager(pyautogit.screen_manager.ScreenManager):
    """Class representing the manager for the repo select screen

    Attributes
    ----------
    menu_choices : list of str
        Overriden attribute from base class with expanded menu choices.
    """

    def __init__(self, top_manager):
        """Constructor for repo select manager
        """
        
        super().__init__(top_manager, 'repo selection')
        self.menu_choices = ['(Re)Enter Credentials',
                                'Open Directory',
                                'Clone New Repository',
                                'Create New Repository',
                                'Settings',
                                'Enter Custom Command',
                                'Exit']


    def process_menu_selection(self, selection):
        """Override of base class, executes depending on menu selection


        Parameters
        ----------
        selection : str
            The user's menu selection
        """

        if selection == '(Re)Enter Credentials':
            self.manager.ask_credentials()
        elif selection == 'Open Directory':
            # This will be implemented once py_cui adds a filemanager popup.
            self.manager.open_not_supported_popup(selection)
            pass
        elif selection == 'Clone New Repository':
            self.manager.root.move_focus(self.clone_new_box)
        elif selection == 'Create New Repository':
            self.manager.root.move_focus(self.create_new_box)
        elif selection == 'Settings':
            self.manager.open_settings_window()
        elif selection == 'Enter Custom Command':
            self.ask_custom_command()
        elif selection == 'Exit':
            self.manager.close_cleanup()
            exit()
        else:
            self.manager.open_not_supported_popup(selection)


    def initialize_screen_elements(self):
        """Override of base function. Initializes widgets, returns screen widget set

        Returns
        -------
        repo_select_widget_set : py_cui.widget_set.WidgetSet
            Widget set object for repo select screen
        """

        repo_select_widget_set = self.manager.root.create_new_widget_set(5,4)
        
        logo_label = repo_select_widget_set.add_block_label(self.manager.get_logo_text(), 0, 0, column_span=2, center=False)
        logo_label.set_color(py_cui.MAGENTA_ON_BLACK)

        link_label = repo_select_widget_set.add_label('v{} - https://github.com/jwlodek/pyautogit'.format(pyautogit.__version__), 0, 2, column_span=2)
        link_label.add_text_color_rule('https://.*', py_cui.CYAN_ON_BLACK, 'contains', match_type='regex')
        
        self.repo_menu = repo_select_widget_set.add_scroll_menu('Repos in Workspace', 1, 2, row_span=2)
        self.repo_menu.add_item_list(self.manager.repos)
        self.repo_menu.add_key_command(py_cui.keys.KEY_ENTER,   self.manager.open_autogit_window)
        self.repo_menu.add_key_command(py_cui.keys.KEY_SPACE,   self.show_repo_status)
        self.repo_menu.add_key_command(py_cui.keys.KEY_DELETE,  self.ask_delete_repo)
        self.repo_menu.add_key_command(py_cui.keys.KEY_R_LOWER, self.refresh_status)
        self.repo_menu.add_key_command(py_cui.keys.KEY_M_LOWER, self.show_menu)
        self.repo_menu.add_key_command(py_cui.keys.KEY_Q_LOWER, self.manager.clean_exit)
        self.repo_menu.add_key_command(py_cui.keys.KEY_S_LOWER, self.manager.open_settings_window)
        self.repo_menu.add_key_command(py_cui.keys.KEY_C_LOWER, self.manager.ask_credentials)
        self.repo_menu.add_key_command(py_cui.keys.KEY_E_LOWER, self.manager.ask_default_editor)
        self.repo_menu.set_focus_text('Quit - q | Open - Enter | Status - Space | Menu - m | Refresh - r | Delete - Del | Settings - s | Credentials - c | Editor - e')

        self.git_status_box = repo_select_widget_set.add_text_block('Git Repo Status', 1, 0, row_span=4, column_span=2)
        self.git_status_box.is_selectable = False
        self.git_status_box.add_text_color_rule('Welcome', py_cui.GREEN_ON_BLACK, 'startswith', match_type='line')
        
        self.current_status_box = repo_select_widget_set.add_text_block('Current Status', 1, 3, row_span=2)
        self.current_status_box.is_selectable = False
        
        self.clone_new_box = repo_select_widget_set.add_text_box('Clone Repository - Enter Remote URL', 3, 2, column_span=2)
        self.clone_new_box.add_key_command(py_cui.keys.KEY_ENTER, lambda : self.execute_long_operation('Cloning', self.clone_new_repo, credentials_required=True))
        
        self.create_new_box = repo_select_widget_set.add_text_box('Create New Repository - Enter Directory Name', 4, 2, column_span=2)
        self.create_new_box.add_key_command(py_cui.keys.KEY_ENTER, self.create_new_repo)
        
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_S_LOWER, self.manager.open_settings_window)
        #repo_select_widget_set.add_key_command(py_cui.keys.KEY_F_LOWER, self.sync_all_repos)
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_R_LOWER, self.refresh_status)
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_C_LOWER, self.manager.ask_credentials)
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_M_LOWER, self.show_menu)
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_E_LOWER, self.manager.ask_default_editor)
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_A_LOWER, lambda : self.git_status_box.set_text(self.manager.get_about_info()))
        repo_select_widget_set.add_key_command(py_cui.keys.KEY_H_LOWER, lambda : self.git_status_box.set_text(self.manager.get_welcome_message()))

        self.info_panel = self.git_status_box

        self.refresh_status()
        return repo_select_widget_set


    def clear_elements(self):
        """Override of base class function, clears text fields
        """

        self.git_status_box.clear()
        self.clone_new_box.clear()
        self.create_new_box.clear()


    def set_initial_values(self):
        """Override of base function. Sets some initial text for the widgets
        """

        if self.manager.metadata_manager.first_time:
            self.git_status_box.set_text(self.manager.get_welcome_message())
            self.manager.metadata_manager.first_time = False
        else:
            self.git_status_box.set_text(self.manager.get_about_info(with_logo = False))
        self.manager.root.set_status_bar_text('Quit - q | Full Menu - m | Refresh - r | Update Credentials - c | Settings Menu - s')


    def refresh_status(self):
        """Function that refreshes the repositories in the selection screen
        """

        LOGGER.write('Refreshing repo select status')
        self.manager.repos = pyautogit.find_repos_in_path(self.manager.workspace_path)
        self.repo_menu.clear()
        self.repo_menu.add_item_list(self.manager.repos)

        status_message = 'Current directory:\n{}\n\n'.format(self.manager.workspace_path)
        status_message = status_message + '# of Repos: {}\n\n'.format(len(self.manager.repos))
        if len(self.manager.credentials) == 0:
            status_message = status_message + 'Credentials Not Entered\n'
        else:
            status_message = status_message + 'Credentials Entered\n'
        if self.manager.default_editor is not None:
            status_message = status_message + '\nEditor: {}'.format(self.manager.default_editor)
        else:
            status_message = status_message + '\nNo Editor Specified.'
        self.current_status_box.set_text(status_message)


    def ask_delete_repo(self):
        """Function that asks user for confirmation for repo deletion
        """

        target = self.repo_menu.get()
        self.manager.root.show_yes_no_popup("Are you sure you want to delete {}?".format(target), self.delete_repo)


    def delete_repo(self, to_delete):
        """Function that deletes a repo

        Parameters
        ----------
        to_delete : bool
            User's response of request for confirmation of deletion
        """

        if to_delete:
            target = self.repo_menu.get()
            LOGGER.write('Deleting repository {}'.format(target))
            pyautogit.commands.remove_repo_tree(target)
            self.refresh_status()


    def show_repo_status(self):
        """Function that shows the current repository status
        """

        current_repo = self.repo_menu.get_selected_item_index()
        repo_name = self.repo_menu.get()
        LOGGER.write('Displaying repo status for {}'.format(repo_name))
        self.git_status_box.clear()
        out, err = pyautogit.commands.git_status(repo_name)
        if err != 0:
            self.manager.root.show_error_popup('Unable to get git status!', out)
        self.git_status_box.title = 'Git Repo Status - {}'.format(repo_name)
        self.git_status_box.set_text('\n{}'.format(out))
        self.refresh_status()
        self.repo_menu.selected_item = current_repo


    def clone_new_repo(self):
        """Function that clones new repo from given URL
        """

        new_repo_url = self.clone_new_box.get()
        LOGGER.write('Cloning new repo {}'.format(new_repo_url))
        self.message, self.status = pyautogit.commands.git_clone_new_repo(new_repo_url, self.manager.credentials)
        self.refresh_status()
        self.clone_new_box.clear()
        # Turn off loading popup
        self.manager.root.stop_loading_popup()


    def create_new_repo(self):
        """Function that creates a new repo with a given name
        """

        new_dir_target = self.create_new_box.get()
        out, err = pyautogit.commands.git_init_new_repo(new_dir_target)
        if err != 0:
            self.manager.root.show_error_popup('Unable to create new repository!', out)
        else:
            self.manager.root.show_message_popup('Created new repository', out)
        self.create_new_box.clear()
        self.refresh_status()

